import codecs
import decimal
import re

import pytest

from hypothesis import (
    given,
    settings,
    example,
    strategies as st,
)

from eth_utils import (
    int_to_big_endian,
    is_boolean,
    is_integer,
    is_number,
    is_address,
    is_bytes,
    is_text,
    to_normalized_address,
    to_canonical_address,
    to_checksum_address,
    decode_hex,
)

from eth_abi.exceptions import (
    EncodingTypeError,
    IllegalValue,
    ValueOutOfBounds,
)
from eth_abi.encoding import (
    BooleanEncoder,
    UnsignedIntegerEncoder,
    SignedIntegerEncoder,
    AddressEncoder,
    BytesEncoder,
    ByteStringEncoder,
    TextStringEncoder,
    encode_uint_256,
    UnsignedFixedEncoder,
    SignedFixedEncoder,
    UnsignedRealEncoder,
    SignedRealEncoder,
    TupleEncoder,
)

from eth_abi.utils.numeric import (
    TEN,
    abi_decimal_context,
    compute_unsigned_integer_bounds,
    compute_signed_integer_bounds,
    compute_unsigned_fixed_bounds,
    compute_signed_fixed_bounds,
    compute_unsigned_real_bounds,
    compute_signed_real_bounds,
    ceil32,
)
from eth_abi.utils.padding import (
    fpad,
    zpad,
    zpad_right,
)


@given(
    bool_value=st.one_of(st.booleans(), st.none()),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_boolean(bool_value, data_byte_size):
    encoder = BooleanEncoder(
        data_byte_size=data_byte_size,
    )


    if not is_boolean(bool_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(bool_value)
        assert 'BooleanEncoder' in str(exception_info.value)
        return

    expected_value = zpad(b'\x01' if bool_value else b'\x00', data_byte_size)
    encoded_value = encoder(bool_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@example(integer_value=-1, value_bit_size=8, data_byte_size=1)
@given(
    integer_value=st.one_of(st.integers(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_unsigned_integer(integer_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError) as exception_info:
            UnsignedIntegerEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = UnsignedIntegerEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )
    lower_bound, upper_bound = compute_unsigned_integer_bounds(value_bit_size)

    if not is_integer(integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(integer_value)
        assert 'UnsignedInteger' in str(exception_info.value)
        return
    elif integer_value < lower_bound or integer_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(integer_value)
        return

    if integer_value >= 0:
        expected_value = zpad(int_to_big_endian(integer_value), data_byte_size)
    else:
        expected_value = fpad(int_to_big_endian(integer_value), data_byte_size)

    encoded_value = encoder(integer_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    integer_value=st.one_of(st.integers(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_signed_integer(integer_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedIntegerEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = SignedIntegerEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    lower_bound, upper_bound = compute_signed_integer_bounds(value_bit_size)

    if not is_integer(integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(integer_value)
        assert 'SignedInteger' in str(exception_info.value)
        return
    elif integer_value < lower_bound or integer_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(integer_value)
        return

    if integer_value >= 0:
        expected_value = zpad(
            int_to_big_endian(integer_value % 2**value_bit_size),
            data_byte_size
        )
    else:
        expected_value = fpad(
            int_to_big_endian(integer_value % 2**value_bit_size),
            data_byte_size
        )

    encoded_value = encoder(integer_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    address_value=st.one_of(
        st.none(),
        st.binary(min_size=20, max_size=20),
        st.binary(min_size=20, max_size=20).map(to_normalized_address),
        st.binary(min_size=20, max_size=20).map(to_checksum_address),
    ),
    value_bit_size=st.integers(min_value=19, max_value=21).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=18, max_value=32),
)
def test_encode_address(address_value, value_bit_size, data_byte_size):
    if value_bit_size != 20 * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = AddressEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_address(address_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(address_value)
        assert 'AddressEncoder' in str(exception_info.value)
        return

    expected_value = zpad(to_canonical_address(address_value), data_byte_size)
    encoded_value = encoder(address_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    bytes_value=st.one_of(
        st.none(),
        st.binary(min_size=0, max_size=40),
    ),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=18, max_value=32),
)
def test_encode_bytes_xx(bytes_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = BytesEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_bytes(bytes_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(bytes_value)
        assert 'BytesEncoder' in str(exception_info.value)
        return
    elif len(bytes_value) * 8 > value_bit_size:
        with pytest.raises(ValueOutOfBounds):
            encoder(bytes_value)
        return

    expected_value = zpad_right(bytes_value, data_byte_size)
    encoded_value = encoder(bytes_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    string_value=st.one_of(
        st.none(),
        st.text(min_size=0, max_size=256),
        st.binary(min_size=0, max_size=256),
    ),
)
def test_encode_byte_string(string_value):
    encoder = ByteStringEncoder()

    if not is_bytes(string_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(string_value)
        assert 'ByteStringEncoder' in str(exception_info.value)
        return

    expected_value = (
        encode_uint_256(len(string_value)) +
        (
            zpad_right(string_value, ceil32(len(string_value)))
            if string_value
            else b'\x00' * 32
        )
    )
    encoded_value = encoder(string_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    string_value=st.one_of(
        st.none(),
        st.text(min_size=0, max_size=256),
        st.binary(min_size=0, max_size=256),
    ),
)
def test_encode_text_string(string_value):
    encoder = TextStringEncoder()

    if not is_text(string_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(string_value)
        assert 'TextStringEncoder' in str(exception_info.value)
        return

    string_value_as_bytes = codecs.encode(string_value, 'utf8')

    expected_value = (
        encode_uint_256(len(string_value_as_bytes)) +
        (
            zpad_right(
                string_value_as_bytes,
                ceil32(len(string_value_as_bytes)),
            )
            if string_value
            else b'\x00' * 32
        )
    )
    encoded_value = encoder(string_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    base_integer_value=st.one_of(st.integers(), st.none()),
    high_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    low_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
@example(
    base_integer_value=384,
    value_bit_size=232,
    high_bit_size=64,
    low_bit_size=168,
    data_byte_size=29,
)
def test_encode_unsigned_real(base_integer_value,
                              value_bit_size,
                              high_bit_size,
                              low_bit_size,
                              data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedRealEncoder(
                value_bit_size=value_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif high_bit_size + low_bit_size != value_bit_size:
        with pytest.raises(ValueError):
            UnsignedRealEncoder(
                value_bit_size=value_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = UnsignedRealEncoder(
        value_bit_size=value_bit_size,
        high_bit_size=high_bit_size,
        low_bit_size=low_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_number(base_integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(base_integer_value)
        assert 'UnsignedReal' in str(exception_info.value)
        return

    with decimal.localcontext(abi_decimal_context):
        real_value = decimal.Decimal(base_integer_value) / 2 ** low_bit_size

    lower_bound, upper_bound = compute_unsigned_real_bounds(
        high_bit_size,
        low_bit_size,
    )

    if real_value < lower_bound or real_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(base_integer_value)
        return

    expected_value = zpad(int_to_big_endian(base_integer_value), data_byte_size)
    encoded_value = encoder(real_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    base_integer_value=st.one_of(st.integers(), st.none()),
    high_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    low_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_signed_real(base_integer_value,
                              value_bit_size,
                              high_bit_size,
                              low_bit_size,
                              data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedRealEncoder(
                value_bit_size=value_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif high_bit_size + low_bit_size != value_bit_size:
        with pytest.raises(ValueError):
            SignedRealEncoder(
                value_bit_size=value_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = SignedRealEncoder(
        value_bit_size=value_bit_size,
        high_bit_size=high_bit_size,
        low_bit_size=low_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_number(base_integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(base_integer_value)
        assert 'SignedReal' in str(exception_info.value)
        return

    unsigned_integer_value = base_integer_value % 2**(high_bit_size + low_bit_size)

    with decimal.localcontext(abi_decimal_context):
        real_value = decimal.Decimal(unsigned_integer_value) / 2 ** low_bit_size

    lower_bound, upper_bound = compute_signed_real_bounds(
        high_bit_size,
        low_bit_size,
    )

    if real_value < lower_bound or real_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(real_value)
        return

    expected_value = zpad(int_to_big_endian(unsigned_integer_value), data_byte_size)
    encoded_value = encoder(real_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    value=st.one_of(st.integers(), st.decimals(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
@example(value=decimal.Decimal('5.33'), value_bit_size=8, frac_places=1, data_byte_size=1)
def test_encode_unsigned_fixed(value,
                               value_bit_size,
                               frac_places,
                               data_byte_size):
    if value_bit_size > data_byte_size * 8:
        pattern = r'Value byte size exceeds data size'
        with pytest.raises(ValueError, match=pattern):
            UnsignedFixedEncoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    encoder = UnsignedFixedEncoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    if not is_number(value):
        pattern = r'Value of type .*NoneType.* cannot be encoded by UnsignedFixedEncoder'
        with pytest.raises(EncodingTypeError, match=pattern):
            encoder(value)
        return

    if UnsignedFixedEncoder.illegal_value_fn(value):
        pattern = r'Value .*(NaN|Infinity|-Infinity).* cannot be encoded by UnsignedFixedEncoder'
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    lower, upper = compute_unsigned_fixed_bounds(value_bit_size, frac_places)
    if value < lower or value > upper:
        pattern = r'Value .* cannot be encoded in .* bits'
        with pytest.raises(ValueOutOfBounds, match=pattern):
            encoder(value)
        return

    with decimal.localcontext(abi_decimal_context):
        residue = value % (TEN ** -frac_places)
    if residue > 0:
        pattern = re.escape(
            'UnsignedFixedEncoder cannot encode value {}: '
            'residue {} outside allowed fractional precision of {}'.format(
                repr(value),
                repr(residue),
                frac_places,
            )
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    # Ensure no exception
    encoder(value)


@settings(max_examples=250)
@given(
    value=st.one_of(st.integers(), st.decimals(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_encode_signed_fixed(value,
                             value_bit_size,
                             frac_places,
                             data_byte_size):
    if value_bit_size > data_byte_size * 8:
        pattern = r'Value byte size exceeds data size'
        with pytest.raises(ValueError, match=pattern):
            SignedFixedEncoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    encoder = SignedFixedEncoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    if not is_number(value):
        pattern = r'Value of type .*NoneType.* cannot be encoded by SignedFixedEncoder'
        with pytest.raises(EncodingTypeError, match=pattern):
            encoder(value)
        return

    if SignedFixedEncoder.illegal_value_fn(value):
        pattern = r'Value .*(NaN|Infinity|-Infinity).* cannot be encoded by SignedFixedEncoder'
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    lower, upper = compute_signed_fixed_bounds(value_bit_size, frac_places)
    if value < lower or value > upper:
        pattern = r'Value .* cannot be encoded in .* bits'
        with pytest.raises(ValueOutOfBounds, match=pattern):
            encoder(value)
        return

    with decimal.localcontext(abi_decimal_context):
        residue = value % (TEN ** -frac_places)
    if residue > 0:
        pattern = re.escape(
            'SignedFixedEncoder cannot encode value {}: '
            'residue {} outside allowed fractional precision of {}'.format(
                repr(value),
                repr(residue),
                frac_places,
            )
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    # Ensure no exception
    encoder(value)


# TODO: make this generic
def test_tuple_encoder():
    encoder = TupleEncoder(encoders=(
        UnsignedIntegerEncoder(value_bit_size=256),
        ByteStringEncoder(),
    ))
    expected = decode_hex('0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000004000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000')
    actual = encoder((0, b''))
    assert actual == expected



import pytest

from eth_abi import (
    decoding,
    encoding,
    exceptions,
)

from eth_abi.registry import (
    ABIRegistry,
    BaseEquals,
    has_arrlist,
    registry as default_registry,
)


@pytest.fixture
def registry():
    reg = ABIRegistry()

    reg.register(
        has_arrlist,
        encoding.BaseArrayEncoder, decoding.BaseArrayDecoder,
        label='has_arrlist',
    )
    reg.register(
        BaseEquals('address'),
        encoding.AddressEncoder, decoding.AddressDecoder,
        label='address',
    )

    return reg


@pytest.mark.parametrize(
    'type_str, encoder_class, decoder_class',
    (
        ('address', encoding.AddressEncoder, decoding.AddressDecoder),
        ('bool', encoding.BooleanEncoder, decoding.BooleanDecoder),
        ('bytes12', encoding.BytesEncoder, decoding.BytesDecoder),
        ('function', encoding.BytesEncoder, decoding.BytesDecoder),
        ('bytes', encoding.ByteStringEncoder, decoding.ByteStringDecoder),
        ('int', encoding.SignedIntegerEncoder, decoding.SignedIntegerDecoder),
        ('int128', encoding.SignedIntegerEncoder, decoding.SignedIntegerDecoder),
        ('string', encoding.TextStringEncoder, decoding.StringDecoder),
        ('uint', encoding.UnsignedIntegerEncoder, decoding.UnsignedIntegerDecoder),
        ('uint8', encoding.UnsignedIntegerEncoder, decoding.UnsignedIntegerDecoder),
        ('ureal128x128', encoding.UnsignedRealEncoder, decoding.UnsignedRealDecoder),
        ('real128x128', encoding.SignedRealEncoder, decoding.SignedRealDecoder),
        ('int[]', encoding.DynamicArrayEncoder, decoding.DynamicArrayDecoder),
        ('int[2]', encoding.SizedArrayEncoder, decoding.SizedArrayDecoder),
        ('int[2][]', encoding.DynamicArrayEncoder, decoding.DynamicArrayDecoder),
        ('int[][2]', encoding.SizedArrayEncoder, decoding.SizedArrayDecoder),
    ),
)
def test_default_registry_gives_correct_basic_coders(type_str, encoder_class, decoder_class):
    assert isinstance(default_registry.get_encoder(type_str), encoder_class)
    assert isinstance(default_registry.get_decoder(type_str), decoder_class)


def test_cache_resets_after_register_and_register_works(registry: ABIRegistry):
    # Populate cache
    registry.get_encoder('address')
    registry.get_decoder('address')

    # Perform cache resetting action
    registry.register(
        BaseEquals('address', with_sub=False),
        encoding.AddressEncoder, decoding.AddressDecoder,
        label='address without sub',
    )

    # Confirm cache reset
    pattern = r'Multiple matching entries .* encoder registry'
    with pytest.raises(ValueError, match=pattern):
        registry.get_encoder('address')

    pattern = r'Multiple matching entries .* decoder registry'
    with pytest.raises(ValueError, match=pattern):
        registry.get_decoder('address')


def test_cache_resets_after_unregister_and_unregister_works(registry: ABIRegistry):
    # Populate cache
    registry.get_encoder('address')
    registry.get_decoder('address')

    # Perform cache resetting action
    registry.unregister('address')

    # Confirm cache reset
    with pytest.raises(ValueError, match=r'No matching entries .* encoder registry'):
        registry.get_encoder('address')
    with pytest.raises(ValueError, match=r'No matching entries .* decoder registry'):
        registry.get_decoder('address')


def test_can_register_and_unregister_string_lookups(registry: ABIRegistry):
    registry.register(
        'bool',
        encoding.BooleanEncoder,
        decoding.BooleanDecoder,
    )

    assert isinstance(registry.get_encoder('bool'), encoding.BooleanEncoder)
    assert isinstance(registry.get_decoder('bool'), decoding.BooleanDecoder)

    registry.unregister('bool')

    with pytest.raises(ValueError, match=r'No matching entries .* encoder registry'):
        registry.get_encoder('bool')
    with pytest.raises(ValueError, match=r'No matching entries .* decoder registry'):
        registry.get_decoder('bool')


def test_registry_should_reject_unknown_types(registry: ABIRegistry):
    with pytest.raises(TypeError):
        registry.register(True, None, None)
    with pytest.raises(TypeError):
        registry.unregister(True)


def test_can_unregister_by_equality(registry: ABIRegistry):
    registry.unregister(BaseEquals('address'))

    with pytest.raises(ValueError, match=r'No matching entries .* encoder registry'):
        registry.get_encoder('address')
    with pytest.raises(ValueError, match=r'No matching entries .* decoder registry'):
        registry.get_decoder('address')


def test_can_register_simple_callables(registry: ABIRegistry):
    encode_bool = lambda x: x
    decode_bool = lambda x: x

    registry.register('bool', encode_bool, decode_bool)

    assert registry.get_encoder('bool') is encode_bool
    assert registry.get_decoder('bool') is decode_bool


def test_unregister_unknown_lookups(registry: ABIRegistry):
    with pytest.raises(KeyError, match=r'Matcher .* not found in encoder registry'):
        registry.unregister(lambda x: x)
    with pytest.raises(KeyError, match=r'Label .* not found in encoder registry'):
        registry.unregister('foo')


def test_looking_up_unparsable_type_causes_error(registry: ABIRegistry):
    with pytest.raises(exceptions.ParseError, match='Parse error at'):
        registry.get_encoder('uint[]256')



import pytest

from eth_abi.utils.padding import zpad


@pytest.mark.parametrize(
    'value,length,expected',
    (
        (b'', 5, b"\x00\x00\x00\x00\x00"),
        (b'abc', 5, b"\x00\x00abc"),
        (b'abcde', 5, b"abcde"),
        (b'abcdef', 5, b"abcdef"),
    )
)
def test_zpadding(value, length, expected):
    actual = zpad(value, length)
    assert actual == expected



import pytest

from eth_abi.exceptions import ParseError

from eth_abi.utils.parsing import (
    collapse_type,
    process_type,
)


@pytest.mark.parametrize(
    'typestr, expected_parse',
    (
        ('uint256', ('uint', '256', [])),
        ('uint', ('uint', '256', [])),
        ('uint256[]', ('uint', '256', [[]])),
        ('uint[]', ('uint', '256', [[]])),
        ('int[2]', ('int', '256', [[2]])),
        ('function', ('bytes', '24', [])),
        ('fixed', ('fixed', '128x18', [])),
        ('ufixed', ('ufixed', '128x18', [])),
    )
)
def test_process_type(typestr, expected_parse):
    assert process_type(typestr) == expected_parse


def test_process_type_requires_string_type():
    with pytest.raises(TypeError):
        process_type(b'uint256')


def test_process_type_requires_basic_type():
    with pytest.raises(ValueError):
        process_type('(int,int)')


@pytest.mark.parametrize(
    'typestr',
    (
        'fixed264x1',
        'fixed9x1',
        'fixed256x81',
        'ufixed264x1',
        'ufixed9x1',
        'ufixed256x81',
    )
)
def test_process_validation_errors(typestr):
    with pytest.raises(ValueError):
        process_type(typestr)


@pytest.mark.parametrize(
    'typestr',
    (
        'fixed0x1',
        'fixed256x0',
        'ufixed0x1',
        'ufixed256x0',
    )
)
def test_process_parsing_errors(typestr):
    with pytest.raises(ParseError):
        process_type(typestr)


@pytest.mark.parametrize(
    'original, expected',
    [
        ('address', 'address'),
        ('uint[2][]', 'uint256[2][]'),
        ('uint256[2][]', 'uint256[2][]'),
        ('function', 'bytes24'),
        ('bool', 'bool'),
        ('bytes32', 'bytes32'),
        ('bytes', 'bytes'),
        ('string', 'string'),
    ],
)
def test_collapse_type(original, expected):
    assert collapse_type(*process_type(original)) == expected



import pytest

from eth_abi.abi import encode_single

from ..common.unit import CORRECT_SINGLE_ENCODINGS


@pytest.mark.parametrize('typ,python_value,expected', CORRECT_SINGLE_ENCODINGS)
def test_encode_single(typ, python_value, expected):
    actual = encode_single(typ, python_value)
    assert actual == expected



import functools
import operator
import pytest

from eth_abi.registry import (
    PredicateMapping,
    Equals,
)


@pytest.fixture
def mapping():
    return PredicateMapping('registry')


equals_foo = Equals('foo')
equals_bar = Equals('bar')


def test_can_add_predicates_and_find_corresponding_values(mapping):
    mapping.add(equals_foo, 'foo value')
    mapping.add(equals_bar, 'bar value')

    assert mapping.find('foo') == 'foo value'
    assert mapping.find('bar') == 'bar value'


def test_cannot_add_same_predicate_twice(mapping):
    mapping.add(equals_foo, 'foo value')

    pattern = r'Matcher .* already exists in registry'
    with pytest.raises(ValueError, match=pattern):
        mapping.add(equals_foo, 'foo value')


def test_cannot_add_two_predicates_with_same_label(mapping):
    mapping.add(equals_foo, 'foo value', 'some label')

    pattern = r"Matcher .* with label 'some label' already exists in registry"
    with pytest.raises(ValueError, match=pattern):
        mapping.add(equals_bar, 'bar value', 'some label')


def test_no_matching_predicates_should_raise_error(mapping):
    mapping.add(equals_foo, 'foo value')
    mapping.add(equals_bar, 'bar value')

    with pytest.raises(ValueError, match=r"No matching entries for 'baz' in registry"):
        mapping.find('baz')


def test_multiple_matching_predicates_should_raise_error(mapping):
    mapping.add(Equals('baz'), 'baz value')
    mapping.add(functools.partial(operator.eq, 'baz'), 'baz value')

    with pytest.raises(ValueError) as e:
        mapping.find('baz')

    e.match(r"Multiple matching entries for 'baz' in registry")
    e.match(r"<Equals \(== 'baz'\)>")
    e.match(r'functools\.partial')


def test_can_remove_predicates_by_reference(mapping):
    mapping.add(equals_foo, 'foo value')
    assert mapping.find('foo') == 'foo value'

    mapping.remove(equals_foo)
    with pytest.raises(ValueError, match=r'No matching'):
        mapping.find('foo')


def test_removing_non_existent_predicate_raises_error(mapping):
    with pytest.raises(KeyError, match=r'Matcher .* not found in registry'):
        mapping.remove(equals_foo)


def test_removing_labeled_predicate_by_reference_also_removes_label(mapping):
    mapping.add(equals_foo, 'foo value', 'foo label')
    mapping.remove(equals_foo)

    # Just a bit of white-box testing here
    assert len(mapping._values) == 0
    assert len(mapping._labeled_predicates) == 0


def test_can_remove_predicates_by_label_if_label_was_given(mapping):
    mapping.add(equals_foo, 'foo value', 'foo label')
    assert mapping.find('foo') == 'foo value'

    mapping.remove('foo label')
    with pytest.raises(ValueError, match=r'No matching'):
        mapping.find('foo')


def test_removing_non_existent_label_raises_error(mapping):
    with pytest.raises(KeyError, match=r"Label 'asdf' not found in registry"):
        mapping.remove('asdf')


def test_removing_unsupported_type_raises_error(mapping):
    with pytest.raises(TypeError, match=r'must be callable or string.*got <class \'bool\'>'):
        mapping.remove(True)



import pytest

from eth_abi.abi import decode_abi

from eth_abi.exceptions import DecodingError

from eth_abi.grammar import parse

from ..common.unit import CORRECT_TUPLE_ENCODINGS


@pytest.mark.parametrize('type_str,expected,byte_str', CORRECT_TUPLE_ENCODINGS)
def test_decode_abi(type_str, expected, byte_str):
    abi_type = parse(type_str)
    if abi_type.arrlist is not None:
        pytest.skip('ABI coding functions do not support array types')

    types = [str(t) for t in abi_type.components]

    actual = decode_abi(types, byte_str)
    assert actual == expected


def test_decode_abi_empty_data_raises():
    with pytest.raises(DecodingError):
        decode_abi(['uint32', 'uint32'], b'')


def test_decode_abi_wrong_data_type_raises():
    with pytest.raises(TypeError):
        decode_abi(['uint32', 'uint32'], '')



import re

from hypothesis import (
    event,
    given,
    settings,
    strategies as st,
)

from eth_abi.registry import (
    Equals,
    BaseEquals,
    has_arrlist,
    is_base_tuple,
)

from tests.common.strategies import (
    malformed_type_strs,
    type_strs,
)


@given(st.text())
def test_equals_has_expected_behavior(s):
    pos_predicate = Equals(s)
    neg_predicate = Equals(s + 'a')

    # Matches correct strings
    assert pos_predicate(s)
    assert not neg_predicate(s)

    # Can be equal/unequal
    assert pos_predicate == Equals(s)
    assert pos_predicate != neg_predicate


def test_equals_can_be_dict_key():
    pred = Equals('foo')
    assert {pred: 'bar'}[pred] == 'bar'


def test_equals_has_expected_repr():
    assert repr(Equals('foo')) == "<Equals (== 'foo')>"


ARRAY_RE = re.compile(r'\[\d*\]')


@settings(max_examples=200)
@given(type_strs)
def test_base_equals_has_expected_behavior_for_parsable_types(type_str):
    is_int = BaseEquals('int')
    is_int_with_sub = BaseEquals('int', with_sub=True)
    is_int_with_no_sub = BaseEquals('int', with_sub=False)

    # Should not match tuple types
    if type_str.startswith('('):
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event('No match for tuple type')

    # Should not match array types
    elif ARRAY_RE.search(type_str):
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event('No match for array type')

    # Should match types with int base
    elif type_str.startswith('int'):
        assert is_int(type_str)
        event('Match for base')

        if type_str == 'int':
            assert is_int_with_no_sub(type_str)
            assert not is_int_with_sub(type_str)
            event('Match for base with no sub')
        else:
            assert not is_int_with_no_sub(type_str)
            assert is_int_with_sub(type_str)
            event('Match for base with sub')

    # Should not match any other types
    else:
        assert not is_int(type_str)
        assert not is_int_with_sub(type_str)
        assert not is_int_with_no_sub(type_str)
        event('No match for other base')


def test_base_equals_can_be_dict_key():
    pred = BaseEquals('foo')
    assert {pred: 'bar'}[pred] == 'bar'


def test_base_equals_has_expected_repr():
    assert repr(BaseEquals('foo')) == "<BaseEquals (base == 'foo')>"
    assert (
        repr(BaseEquals('foo', with_sub=True))
        == "<BaseEquals (base == 'foo' and sub is not None)>"
    )
    assert (
        repr(BaseEquals('foo', with_sub=False))
        == "<BaseEquals (base == 'foo' and sub is None)>"
    )


@given(type_strs)
def test_has_arrlist_has_expected_behavior_for_parsable_types(type_str):
    # Should not match tuple types
    if type_str.startswith('('):
        assert not has_arrlist(type_str)
        event('No match for tuple type')

    # Should match array types
    elif ARRAY_RE.search(type_str):
        assert has_arrlist(type_str)
        event('Match for array type')

    # Should not match any other types
    else:
        assert not has_arrlist(type_str)
        event('No match for non-array type')


@given(type_strs)
def test_is_base_tuple_has_expected_behavior_for_parsable_types(type_str):
    # Should match tuple types
    if type_str.endswith(')'):
        assert is_base_tuple(type_str)
        event('Match for tuple type')

    # Should not match any other types
    else:
        assert not is_base_tuple(type_str)
        event('No match for non-tuple type')


@given(malformed_type_strs)
def test_predicates_have_expected_behavior_for_malformed_types(malformed_type_str):
    is_int = BaseEquals('int')
    is_int_with_sub = BaseEquals('int', with_sub=True)
    is_int_with_no_sub = BaseEquals('int', with_sub=False)

    # Should not match unparsable types
    assert not is_int(malformed_type_str)
    assert not is_int_with_sub(malformed_type_str)
    assert not is_int_with_no_sub(malformed_type_str)

    assert not has_arrlist(malformed_type_str)
    assert not is_base_tuple(malformed_type_str)



import pytest


from eth_abi.base import (
    BaseCoder,
    parse_type_str,
    parse_tuple_type_str,
)

from eth_abi.grammar import BasicType


class Coder(BaseCoder):
    base = None
    sub = None

    def validate(self):
        if self.base is None:
            raise ValueError('Setting missing')

    @parse_type_str(with_arrlist=False)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub)


class ArrayCoder(Coder):
    arrlist = None

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub, arrlist=abi_type.arrlist)


class IntCoder(Coder):
    @parse_type_str('int', with_arrlist=False)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub)


class IntArrayCoder(ArrayCoder):
    @parse_type_str('int', with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        return cls(base=abi_type.base, sub=abi_type.sub, arrlist=abi_type.arrlist)


class TupleCoder(BaseCoder):
    components = None

    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry):
        return cls(components=abi_type.components)


def test_base_coder_requires_that_settings_are_known():
    pattern = 'only accepts keyword arguments which are present'
    with pytest.raises(AttributeError, match=pattern):
        coder = Coder(foo='bar')


def test_base_coder_validates_settings():
    with pytest.raises(ValueError, match='Setting missing'):
        coder = Coder(sub='foo')


def test_base_coder_assigns_all_kwargs_to_instance():
    coder = Coder(base='foo', sub='bar')

    assert coder.base == 'foo'
    assert coder.sub == 'bar'


def test_parse_type_str_normalizes_and_parses_types():
    coder1 = Coder.from_type_str('int', None)

    assert coder1.base == 'int'
    assert coder1.sub == 256


def test_parse_type_str_validates_standard_types():
    with pytest.raises(ValueError, match='integer size must be multiple of 8'):
        Coder.from_type_str('int255', None)


def test_parse_type_str_requires_certain_base_type_if_given():
    with pytest.raises(ValueError, match='expected type with base \'int\''):
        IntCoder.from_type_str('uint256', None)

    with pytest.raises(ValueError, match='expected type with base \'int\''):
        IntArrayCoder.from_type_str('uint256[]', None)


def test_parse_type_str_doesnt_require_base_type_if_not_given():
    coder1 = Coder.from_type_str('uint256', None)
    assert coder1.base == 'uint'
    assert coder1.sub == 256

    coder2 = ArrayCoder.from_type_str('uint256[]', None)
    assert coder2.base == 'uint'
    assert coder2.sub == 256
    assert coder2.arrlist == ((),)


def test_parse_type_str_distinguishes_non_array_and_array_types():
    with pytest.raises(ValueError, match='expected type with no array dimension list'):
        Coder.from_type_str('int256[]', None)
    with pytest.raises(ValueError, match='expected type with array dimension list'):
        ArrayCoder.from_type_str('int256', None)

    with pytest.raises(ValueError, match='expected type with no array dimension list'):
        IntCoder.from_type_str('int256[]', None)
    with pytest.raises(ValueError, match='expected type with array dimension list'):
        IntArrayCoder.from_type_str('int256', None)


def test_parse_tuple_type_str_normalizes_and_parses_types():
    tuple_coder = TupleCoder.from_type_str('(int,int)', None)

    assert tuple_coder.components == (
        BasicType('int', 256, None),
        BasicType('int', 256, None),
    )


def test_parse_tuple_type_str_validates_standard_types():
    with pytest.raises(ValueError, match='integer size must be multiple of 8'):
        TupleCoder.from_type_str('(int255,int)', None)


def test_parse_tuple_type_str_distinguishes_non_tuple_and_tuple_types():
    with pytest.raises(ValueError, match='Cannot create TupleCoder for non-tuple type'):
        TupleCoder.from_type_str('int256', None)



import pytest

from hypothesis import (
    example,
    given,
    strategies as st,
)

from eth_abi.exceptions import (
    ABITypeError,
    ParseError,
)

from eth_abi.grammar import (
    BasicType,
    TYPE_ALIASES,
    TupleType,
    normalize,
    parse,
)

from .common.strategies import (
    malformed_type_strs,
    type_strs,
)


@pytest.mark.parametrize(
    'type_str, expected_type',
    (
        ('uint', BasicType('uint')),
        ('uint256', BasicType('uint', 256)),
        ('uint256[]', BasicType('uint', 256, ((),))),
        ('uint256[][1][23][]', BasicType('uint', 256, ((), (1,), (23,), ()))),
        ('uint[]', BasicType('uint', None, ((),))),
        ('int[2]', BasicType('int', None, ((2,),))),
        ('fixed128x19', BasicType('fixed', (128, 19))),
        ('ufixed128x19[][]', BasicType('ufixed', (128, 19), ((), ()))),
        (
            '(int,ufixed128x19[][])',
            TupleType((
                BasicType('int'),
                BasicType('ufixed', (128, 19), ((), ())),
            )),
        ),
        (
            '(int,(address,function))',
            TupleType((
                BasicType('int'),
                TupleType((
                    BasicType('address'),
                    BasicType('function'),
                )),
            )),
        ),
        (
            '((address,(fixed,bool)),(uint256[][1][],fixed128x19))',
            TupleType((
                TupleType((
                    BasicType('address'),
                    TupleType((
                        BasicType('fixed'),
                        BasicType('bool'),
                    )),
                )),
                TupleType((
                    BasicType('uint', 256, ((), (1,), ())),
                    BasicType('fixed', (128, 19)),
                )),
            )),
        ),
        ('()', TupleType(tuple())),
        (
            '(int,(),())',
            TupleType((BasicType('int'), TupleType(tuple()), TupleType(tuple()))),
        ),
    )
)
def test_parsing_with_parsimonious_grammar_and_node_visitor_works(type_str, expected_type):
    assert parse(type_str) == expected_type


@given(malformed_type_strs, st.none())
@example('int0', 4)
@example('fixed128x0', 9)
@example('ufixed128x0', 10)
@example('int!', 4)
@example('(int, bool)', 6)
@example('(int,bool,)', 11)
@example('(int,(address,uint256[][])', 27)
def test_parsing_invalid_type_str_causes_parse_error(type_str, error_col):
    if error_col is not None:
        pattern = r'Parse error at .* \(column {}\)'.format(error_col)
    else:
        pattern = r'Parse error at .*'

    with pytest.raises(ParseError, match=pattern):
        parse(type_str)


def test_parse_raises_type_error_for_wrong_data_type():
    with pytest.raises(TypeError):
        parse(b'uint256')


@given(type_strs)
def test_end_to_end_parsing_and_collapsing(type_str):
    assert str(parse(type_str)) == type_str


@pytest.mark.parametrize(
    'type_str, pattern',
    (
        ('string1x1', 'string type cannot have suffix'),
        ('string33', 'string type cannot have suffix'),
        ('bytes1x1', 'no suffix or a numerical'),
        ('bytes33', 'maximum 32 bytes'),
        ('int', 'must have numerical suffix'),
        ('int7', 'size out of bounds'),
        ('int257', 'size out of bounds'),
        ('int9', 'must be multiple of 8'),
        ('uint', 'must have numerical suffix'),
        ('uint7', 'size out of bounds'),
        ('uint257', 'size out of bounds'),
        ('uint9', 'must be multiple of 8'),
        ('fixed', 'must have suffix of form'),
        ('fixed8', 'must have suffix of form'),
        ('fixed7x19', 'fixed size out of bounds'),
        ('fixed257x19', 'fixed size out of bounds'),
        ('fixed17x19', 'must be multiple of 8'),
        ('fixed128x81', 'exponent size out of bounds'),
        ('ufixed', 'must have suffix of form'),
        ('ufixed8', 'must have suffix of form'),
        ('ufixed7x19', 'fixed size out of bounds'),
        ('ufixed257x19', 'fixed size out of bounds'),
        ('ufixed17x19', 'must be multiple of 8'),
        ('ufixed128x81', 'exponent size out of bounds'),
        ('real', 'must have suffix of form'),
        ('real8', 'must have suffix of form'),
        ('real6x1', 'real size out of bounds'),
        ('real256x1', 'real size out of bounds'),
        ('real17x8', 'must be multiples of 8'),
        ('real16x9', 'must be multiples of 8'),
        ('ureal', 'must have suffix of form'),
        ('ureal8', 'must have suffix of form'),
        ('ureal6x1', 'real size out of bounds'),
        ('ureal256x1', 'real size out of bounds'),
        ('ureal17x8', 'must be multiples of 8'),
        ('ureal16x9', 'must be multiples of 8'),
        ('hash', 'must have numerical suffix'),
        ('hash128x128', 'must have numerical suffix'),
        ('address128', 'cannot have suffix'),
        ('address128x128', 'cannot have suffix'),
        (
            '(int256,(address,(ufixed7x19,fixed)))',
            "For 'ufixed7x19' type at column 19 "
            r"in '\(int256,\(address,\(ufixed7x19,fixed\)\)\)': "
            "fixed size out of bounds",
        ),
        (
            '(int256,(address,(ufixed128x19,fixed)))',
            "For 'fixed' type at column 32 "
            r"in '\(int256,\(address,\(ufixed128x19,fixed\)\)\)': "
            "fixed type must have suffix of form",
        ),
    )
)
def test_invalid_abi_types(type_str, pattern):
    with pytest.raises(ABITypeError, match=pattern):
        parse(type_str).validate()


@pytest.mark.parametrize(
    'type_str',
    (
        'string',
        'bytes24',
        'int256',
        'uint128',
        'fixed128x19',
        'ufixed256x19',
        'real128x128',
        'ureal128x64',
        'hash256',
        'address',
        '(int256,(address,(ufixed128x19,fixed128x19)))',
    )
)
def test_valid_abi_types(type_str):
    parse(type_str).validate()


@pytest.mark.parametrize(
    'type_str, normalized',
    tuple(TYPE_ALIASES.items()) + (
        ('(int,uint,fixed,ufixed)', '(int256,uint256,fixed128x18,ufixed128x18)'),
        ('(function,function,function)', '(bytes24,bytes24,bytes24)'),
    ),
)
def test_normalize(type_str, normalized):
    assert normalize(type_str) == normalized


def test_basic_type_item_type_throws_errors():
    bt = parse('int256')

    pattern = "Cannot determine item type for non-array type 'int256'"
    with pytest.raises(ValueError, match=pattern):
        bt.item_type



from hypothesis import given

from eth_abi import (
    encode_abi,
    decode_abi,
    encode_single,
    decode_single,
)

from tests.common.strategies import (
    uint_strs,
    uint_values,
)


@given(uint_strs, uint_values)
def test_always_positive_multi(_type, value):
    """
    Tests round trip encoding and decoding for basic types and lists of basic
    types.
    """
    assert value >= 0

    encoded_value = encode_abi([_type], [value])
    decoded_value = decode_abi([_type], encoded_value)

    assert decoded_value[0] >= 0


@given(uint_strs, uint_values)
def test_always_positive_single(_type, value):
    """
    Tests round trip encoding and decoding for basic types and lists of basic
    types.
    """
    assert value >= 0

    encoded_value = encode_single(_type, value)
    decoded_value = decode_single(_type, encoded_value)

    assert decoded_value >= 0



from decimal import Decimal

from eth_utils import decode_hex

from eth_abi.utils.padding import (
    zpad32,
    zpad32_right,
    fpad32,
)


def make_word(description: str) -> bytes:
    r"""
    Converts a "description" of a 32-byte word into a 32-byte word.  A
    description is either a hex representation of a byte string (e.g.
    'deafbeef') or a hex representation along with a fill directive
    'leftpadchar<' or '>rightpadchar'.

    Examples:
    >>> # Left padding examples
    >>> assert make_word('0<deadbeef') == zpad32(b'\xde\xad\xbe\xef')
    >>> assert make_word('f<deadbeef') == fpad32(b'\xde\xad\xbe\xef')
    >>> assert make_word('deadbeef') == zpad32(b'\xde\xad\xbe\xef')
    >>> # Right padding examples
    >>> assert make_word('deadbeef>0') == zpad32_right(b'\xde\xad\xbe\xef')
    """
    if '<' in description:
        fill_char, hex_str = description.split('<')
        return decode_hex(hex_str.rjust(64, fill_char))

    if '>' in description:
        hex_str, fill_char = description.split('>')
        return decode_hex(hex_str.ljust(64, fill_char))

    return decode_hex(description.rjust(64, '0'))


def words(*descriptions: str) -> bytes:
    r"""
    Converts multiple word descriptions into 32-byte words joined into a single
    byte string.

    Examples:
    >>> assert words('1') == zpad32(b'\x01')
    >>> assert words('1', '2f>0') == zpad32(b'\x01') + zpad32_right(b'\x2f')
    """
    return b''.join(make_word(d) for d in descriptions)


CORRECT_TUPLE_ENCODINGS = [
    # (type string, python value, byte string)

    # Empty tuples
    (
        '()', (), b'',
    ),
    (
        '((),((),((),())))',
        ((), ((), ((), ()))),
        b'',
    ),

    # Static tuples
    (
        '(uint32)',
        (6,),
        words('6'),
    ),
    (
        '(uint32,uint32)',
        (2 ** 32 - 1, 2 ** 32 - 1),
        words('ffffffff', 'ffffffff'),
    ),
    (
        '(bytes32,bytes32)',
        (zpad32_right(b'a'), zpad32_right(b'b')),
        words('61>0', '62>0'),
    ),
    (
        '(address,uint32,bytes32,int32)',
        ('0x82a978b3f5962a5b0957d9ee9eef472ee55b42f1', 1, zpad32_right(b'stupid pink animal'), 0),
        words('82a978b3f5962a5b0957d9ee9eef472ee55b42f1', '1', '7374757069642070696e6b20616e696d616c>0', '0'),
    ),

    # Dynamic tuples
    (
        '(bytes32[])',
        ((zpad32_right(b'a'), zpad32_right(b'b')),),
        words('20', '2', '61>0', '62>0'),
    ),
    (
        '(uint256,bytes)',
        (0, b''),
        words('0', '40', '0', '0'),
    ),
    (
        '(int,(int,int[]))',
        (1, (2, (3, 3))),
        words('1', '40', '2', '40', '2', '3', '3')
    ),
    (
        '((int[],int),int)',
        (((1, 1), 2), 3),
        words('40', '3', '40', '2', '2', '1', '1')
    ),
    (
        '((bytes,bool),(bytes,bool))',
        ((b'david attenborough', False), (b'boaty mcboatface', True)),
        words(
            '40',  # offset of first (bytes,bool)
            'c0',  # offset of second (bytes,bool)
            '40',  # offset of bytes in first (bytes,bool)
            '0',  # encoding for `False`
            '12',  # length of b'david attenborough'
            '646176696420617474656e626f726f756768>0',  # encoding of b'david attenborough'
            '40',  # offset of bytes in second (bytes,bool)
            '1',  # encoding for `True`
            '10',  # length of b'boaty mcboatface'
            '626f617479206d63626f617466616365>0',  # encoding of b'boaty mcboatface'
        )
    ),

    # Tuple arrays
    (
        '((int,int)[])',
        (((1, 2), (3, 4)),),
        words('20', '2', '1', '2', '3', '4'),
    ),
    (
        '((int,int[],(int,int)[]),(int,int),int)',
        (
            (
                1,
                (2, 3),
                ((4, 5), (6, 7)),
            ),
            (8, 9),
            10,
        ),
        words(
            '80',  # offset of dynamic tuple
            '8',  # outer tuple static tuple
            '9',
            'a',  # outer tuple int
            '1',  # beginning of dynamic tuple (int)
            '60',  # offset of dynamic tuple list of ints
            'c0',  # offset of dynamic tuple list of tuples
            '2',  # size of list of ints
            '2',
            '3',
            '2',  # size of list of tuples
            '4',  # beginning of first tuple
            '5',
            '6',  # beginning of second tuple
            '7',
        ),
    ),
    (
        '(int,int)[][]',
        (
            ((1, 2),),
            ((3, 4), (5, 6)),
            ((7, 8), (9, 10), (11, 12)),
        ),
        words(
            '3',  # size of outer dynamic list
            '60',  # offset of first dynamic list
            'c0',  # offset of second dynamic list
            '160',  # offset of third dynamic list
            '1',  # size of first list
            '1',  # start of first tuple
            '2',
            '2',  # size of second list
            '3',  # start of second tuple
            '4',
            '5',  # start of third tuple
            '6',
            '3',  # size of third list
            '7',  # start of fourth tuple
            '8',
            '9',  # start of fifth tuple
            'a',
            'b',  # start of sixth tuple
            'c',
        ),
    ),
    (
        '((int,int)[][2])',
        (
            (
                ((1, 2), (3, 4)),
                ((5, 6), (7, 8), (9, 10)),
            ),
        ),
        words(
            '20',  # offset of constant size array
            '40',  # offset of first dynamic list of tuples
            'e0',  # offset of second dynamic list of tuples
            '2',  # size of first list
            '1',  # start of first tuple
            '2',
            '3',  # start of second tuple
            '4',
            '3',  # size of second list
            '5',  # start of third tuple
            '6',
            '7',  # start of fourth tuple
            '8',
            '9',  # start of fifth tuple
            'a',
        ),
    ),
]

CORRECT_SINGLE_ENCODINGS = CORRECT_TUPLE_ENCODINGS + [
    # encode_single/decode_single accept tuple of type components
    (('uint', '256', []), 2 ** 256 - 1, words('f<f')),

    #####
    # (type string, python value, byte string)
    #####

    # uint<M>
    ('uint8', 255, words('ff')),
    ('uint8', 21, words('15')),
    ('uint8', 1, words('1')),
    ('uint256', 2 ** 256 - 1, words('f<f')),
    ('uint256', 2 ** 256 - 100, words('f<9c')),
    ('uint256', 21, words('15')),
    ('uint256', 1, words('1')),

    # int<M>
    ('int8', 21, words('15')),
    ('int8', 1, words('1')),
    ('int8', -1, words('f<f')),
    ('int8', -100, words('f<9c')),
    ('int256', 21, words('15')),
    ('int256', 1, words('1')),
    ('int256', -1, words('f<f')),
    ('int256', -100, words('f<9c')),

    # address
    ('address', '0x0000000000000000000000000000000000000000', words('0')),
    ('address', '0xd3cda913deb6f67967b99d67acdfa1712c293601', words('d3cda913deb6f67967b99d67acdfa1712c293601')),
    ('address', '0x0005c901078781c232a2a521c2af7980f8385ee9', words('0005c901078781c232a2a521c2af7980f8385ee9')),
    ('address', '0x5c901078781c232a2a521c2af7980f8385ee9000', words('5c901078781c232a2a521c2af7980f8385ee9000')),

    # uint, int
    ('uint', 2 ** 256 - 1, words('f<f')),
    ('uint', 2 ** 256 - 100, words('f<9c')),
    ('uint', 21, words('15')),
    ('uint', 1, words('1')),
    ('int', 21, words('15')),
    ('int', 1, words('1')),
    ('int', -1, words('f<f')),
    ('int', -100, words('f<9c')),

    # bool
    ('bool', True, words('1')),
    ('bool', False, words('0')),

    # fixed<M>x<N>
    ('fixed8x1', Decimal('127e-1'), words('7f')),
    ('fixed8x1', Decimal('1e-1'), words('1')),
    ('fixed8x1', Decimal('0'), words('0')),
    ('fixed8x1', Decimal('-1e-1'), words('f<f')),
    ('fixed8x1', Decimal('-128e-1'), words('f<80')),

    ('fixed128x18', Decimal('127e-18'), words('7f')),
    ('fixed128x18', Decimal('1e-18'), words('1')),
    ('fixed128x18', Decimal('0'), words('0')),
    ('fixed128x18', Decimal('-1e-18'), words('f<f')),
    ('fixed128x18', Decimal('-128e-18'), words('f<80')),

    ('fixed256x80', Decimal('127e-80'), words('7f')),
    ('fixed256x80', Decimal('1e-80'), words('1')),
    ('fixed256x80', Decimal('0'), words('0')),
    ('fixed256x80', Decimal('-1e-80'), words('f<f')),
    ('fixed256x80', Decimal('-128e-80'), words('f<80')),

    # ufixed<M>x<N>
    ('ufixed8x1', Decimal('255e-1'), words('ff')),
    ('ufixed8x1', Decimal('254e-1'), words('fe')),
    ('ufixed8x1', Decimal('1e-1'), words('1')),
    ('ufixed8x1', Decimal('0'), words('0')),

    ('ufixed128x18', Decimal('255e-18'), words('ff')),
    ('ufixed128x18', Decimal('254e-18'), words('fe')),
    ('ufixed128x18', Decimal('1e-18'), words('1')),
    ('ufixed128x18', Decimal('0'), words('0')),

    ('ufixed256x80', Decimal('255e-80'), words('ff')),
    ('ufixed256x80', Decimal('254e-80'), words('fe')),
    ('ufixed256x80', Decimal('1e-80'), words('1')),
    ('ufixed256x80', Decimal('0'), words('0')),

    # fixed, ufixed
    ('fixed', Decimal('127e-18'), words('7f')),
    ('fixed', Decimal('1e-18'), words('1')),
    ('fixed', Decimal('0'), words('0')),
    ('fixed', Decimal('-1e-18'), words('f<f')),
    ('fixed', Decimal('-128e-18'), words('f<80')),
    ('ufixed', Decimal('255e-18'), words('ff')),
    ('ufixed', Decimal('254e-18'), words('fe')),
    ('ufixed', Decimal('1e-18'), words('1')),
    ('ufixed', Decimal('0'), words('0')),

    # bytes<M>
    ('bytes32', zpad32_right(b'test'), words('74657374>0')),
    ('bytes32', zpad32_right(b'abcdefghijklmnopqrstuvwxyz'), words('6162636465666768696a6b6c6d6e6f707172737475767778797a>0')),
    ('bytes32', zpad32_right(b'0123456789!@#$%^&*()'), words('3031323334353637383921402324255e262a2829>0')),
    ('bytes32', zpad32_right(b'abc' + 5 * b'\x00' + b'abc'), words('6162630000000000616263>0')),
    ('bytes1', b'a', words('61>0')),

    # bytes
    ('bytes', b'', words('0', '0')),
    ('bytes', b'\xde', words('1', 'de>0')),
]



from decimal import Decimal

from hypothesis import (
    given,
    settings,
)
import pytest

from eth_utils import decode_hex

from eth_abi import is_encodable

from tests.common.strategies import (
    single_strs_values,
    tuple_strs_values,
)
from tests.common.unit import CORRECT_SINGLE_ENCODINGS


@pytest.mark.parametrize(
    'type_str,python_value,_',
    CORRECT_SINGLE_ENCODINGS,
)
def test_is_encodable_returns_true(type_str, python_value, _):
    assert is_encodable(type_str, python_value)


@pytest.mark.parametrize(
    'type_str,python_value',
    (
        # Wrong value type
        ('((bytes,bool),(bytes,bool))', ((b'david attenborough', 0), (b'boaty mcboatface', True))),
        ('(uint32)', ('6',)),
        ('int[]', 6),
        ('bool[2]', (True, 2)),
        ('bool', 1),
        ('uint', True),
        ('int', True),
        ('bytes', 129),
        ('fixed8x1', 0.1),  # only Decimal and int are allowed

        # List size mismatch
        ('int[3]', (6, 2)),

        # Bytes string is wrong size
        ('bytes5', b'asdfgt'),

        # Addresses aren't 20 bytes long
        ('address', '0x8deebe59332ebc3f81935e311f0fb8556cf573'),
        ('address', decode_hex('0x8deebe59332ebc3f81935e311f0fb8556cf573')),

        # Values out of bounds
        ('uint8', 256),
        ('uint8', -1),
        ('int8', 128),
        ('int8', -129),
        ('string', -129),
        ('fixed8x1', Decimal('128e-1')),
        ('fixed8x1', Decimal('-129e-1')),
        ('ufixed8x1', Decimal('256e-1')),
        ('ufixed8x1', Decimal('-1e-1')),
        ('fixed8x1', Decimal('1e-2')),

    )
)
def test_is_encodable_returns_false(type_str, python_value):
    assert not is_encodable(type_str, python_value)


@settings(deadline=None)
@given(single_strs_values)
def test_is_encodable_returns_true_for_random_valid_values(type_and_value):
    _type, value = type_and_value
    assert is_encodable(_type, value)


@settings(deadline=None)
@given(tuple_strs_values)
def test_is_encodable_returns_true_for_random_valid_tuple_values(type_and_value):
    _type, value = type_and_value
    assert is_encodable(_type, value)



import pytest

from eth_abi.abi import decode_single

from ..common.unit import CORRECT_SINGLE_ENCODINGS


@pytest.mark.parametrize('typ,expected,byte_str', CORRECT_SINGLE_ENCODINGS)
def test_decode_single(typ, expected, byte_str):
    actual = decode_single(typ, byte_str)
    assert actual == expected


def test_decode_single_wrong_data_type_raises():
    with pytest.raises(TypeError):
        decode_single('uint32', '')



import pytest

from eth_abi import encode_abi

from eth_abi.grammar import parse

from ..common.unit import CORRECT_TUPLE_ENCODINGS


@pytest.mark.parametrize('type_str,python_value,expected', CORRECT_TUPLE_ENCODINGS)
def test_encode_abi(type_str, python_value, expected):
    abi_type = parse(type_str)
    if abi_type.arrlist is not None:
        pytest.skip('ABI coding functions do not support array types')

    types = [str(t) for t in abi_type.components]

    actual = encode_abi(types, python_value)
    assert actual == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)


setup(
    name='cpc-abi',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='1.3.2',
    description="""CPChain ABI Utils""",
    long_description_content_type="text/markdown",
    long_description_markdown_filename='README.md',
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    url='https://github.com/cpchain/cpc-abi',
    include_package_data=True,
    install_requires=[
        'eth-utils>=1.2.0,<2.0.0',
        'eth-typing>=2.0.0,<3.0.0',
        'parsimonious>=0.8.0,<0.9.0',
    ],
    extras_require={
        'doc': [
            "Sphinx>=1.6.5,<2",
            "sphinx_rtd_theme>=0.1.9",
        ],
    },
    setup_requires=['setuptools-markdown'],
    py_modules=['eth_abi'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={'eth_abi': ['py.typed']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



import decimal
import sys

from eth_utils import (
    big_endian_to_int,
    decode_hex,
    int_to_big_endian,
    to_normalized_address,
)
from eth_utils.toolz import (
    complement,
)
from hypothesis import (
    example,
    given,
    settings,
    strategies as st,
)
import pytest

from eth_abi.constants import (
    TT256M1,
)

from eth_abi.decoding import (
    ContextFramesBytesIO,
    UnsignedIntegerDecoder,
    SignedIntegerDecoder,
    UnsignedRealDecoder,
    SignedRealDecoder,
    UnsignedFixedDecoder,
    SignedFixedDecoder,
    StringDecoder,
    BytesDecoder,
    TupleDecoder,
    BooleanDecoder,
    AddressDecoder,
    DynamicArrayDecoder,
)

from eth_abi.exceptions import (
    InsufficientDataBytes,
    NonEmptyPaddingBytes,
)

from eth_abi.registry import registry

from eth_abi.utils.padding import (
    zpad32,
)
from eth_abi.utils.numeric import (
    abi_decimal_context,
    compute_signed_integer_bounds,
    quantize_value,
    ceil32,
)


def is_non_empty_non_null_byte_string(value):
    return value and big_endian_to_int(value) != 0


def is_valid_padding_bytes(padding_bytes, data_bytes):
    # Empty padding is always valid
    if len(padding_bytes) == 0:
        return True

    leading_data_bit_is_one = (data_bytes[0] & 0b10000000) == 0b10000000

    if leading_data_bit_is_one:
        # All padding bits must be 1
        if padding_bytes.replace(b'\xff', b'') == b'':
            return True
    else:
        # All padding bits must be 0
        if padding_bytes.replace(b'\x00', b'') == b'':
            return True

    return False


def all_bytes_equal(test_bytes, target):
    if sys.version_info.major < 3:
        return all(byte == chr(target) for byte in test_bytes)
    else:
        return all(byte == target for byte in test_bytes)


@settings(max_examples=250)
@given(
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_unsigned_int(integer_bit_size, stream_bytes, data_byte_size):
    if integer_bit_size % 8 != 0:
        with pytest.raises(ValueError):
            UnsignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = UnsignedIntegerDecoder(
            value_bit_size=integer_bit_size,
            data_byte_size=data_byte_size,
        )


    stream = ContextFramesBytesIO(stream_bytes)
    actual_value = big_endian_to_int(stream_bytes[:data_byte_size])

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif actual_value > 2 ** integer_bit_size - 1:
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
@example(8, b'\x00\x80', 2)
@example(8, b'\xff\xff', 2)
def test_decode_signed_int(integer_bit_size, stream_bytes, data_byte_size):
    if integer_bit_size % 8 != 0:
        with pytest.raises(ValueError):
            SignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = SignedIntegerDecoder(
            value_bit_size=integer_bit_size,
            data_byte_size=data_byte_size,
        )


    stream = ContextFramesBytesIO(stream_bytes)

    padding_bytes = data_byte_size - integer_bit_size // 8

    raw_value = big_endian_to_int(stream_bytes[padding_bytes:data_byte_size])
    if raw_value >= 2 ** (integer_bit_size - 1):
        actual_value = raw_value - 2 ** integer_bit_size
    else:
        actual_value = raw_value

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif (
        (actual_value >= 0 and not all_bytes_equal(stream_bytes[:padding_bytes], 0)) or
        (actual_value < 0 and not all_bytes_equal(stream_bytes[:padding_bytes], 255))
    ):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    string_bytes=st.binary(min_size=0, max_size=256),
    pad_size=st.integers(min_value=0, max_value=32),
)
def test_decode_bytes_and_string(string_bytes, pad_size):
    size_bytes = zpad32(int_to_big_endian(len(string_bytes)))
    padded_string_bytes = string_bytes + b'\x00' * pad_size
    stream_bytes = size_bytes + padded_string_bytes
    stream = ContextFramesBytesIO(stream_bytes)

    decoder = StringDecoder()

    if len(padded_string_bytes) < ceil32(len(string_bytes)):
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    decoded_value = decoder(stream)
    assert decoded_value == string_bytes


@settings(max_examples=250)
@given(
    stream_bytes=st.binary(min_size=1, max_size=32),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_decode_boolean(stream_bytes, data_byte_size):
    stream = ContextFramesBytesIO(stream_bytes)

    decoder = BooleanDecoder(data_byte_size=data_byte_size)

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    padding_bytes = stream_bytes[:data_byte_size][:-1]
    if is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    byte_value = stream_bytes[data_byte_size - 1]

    if byte_value in {0, b'\x00'}:
        actual_value = False
    elif byte_value in {1, b'\x01'}:
        actual_value = True
    else:
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    decoded_value = decoder(stream)
    assert decoded_value is actual_value


@settings(max_examples=250)
@given(
    value_byte_size=st.integers(min_value=1, max_value=32),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_bytes_xx(value_byte_size, stream_bytes, data_byte_size):
    if value_byte_size > data_byte_size:
        with pytest.raises(ValueError):
            BytesDecoder(
                value_bit_size=value_byte_size * 8,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = BytesDecoder(
            value_bit_size=value_byte_size * 8,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    actual_value = stream_bytes[:value_byte_size]
    padding_bytes = stream_bytes[value_byte_size:data_byte_size]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    address_bytes=st.binary(min_size=0, max_size=32),
    padding_size=st.integers(min_value=10, max_value=14),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_address(address_bytes, padding_size, data_byte_size):
    stream_bytes = b'\x00' * padding_size + address_bytes
    if data_byte_size < 20:
        with pytest.raises(ValueError):
            AddressDecoder(
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = AddressDecoder(
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    padding_bytes = stream_bytes[:data_byte_size][:-20]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    actual_value = to_normalized_address(stream_bytes[:data_byte_size][-20:])

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    array_size=st.integers(min_value=0, max_value=32),
    array_values=st.lists(st.integers(min_value=0, max_value=TT256M1), min_size=0, max_size=64).map(tuple),
)
def test_decode_array_of_unsigned_integers(array_size, array_values):
    size_bytes = zpad32(int_to_big_endian(array_size))
    values_bytes = b''.join((
        zpad32(int_to_big_endian(v)) for v in array_values
    ))
    stream_bytes = size_bytes + values_bytes

    decoder = DynamicArrayDecoder(
        item_decoder=UnsignedIntegerDecoder(value_bit_size=256),
    )
    stream = ContextFramesBytesIO(stream_bytes)

    if len(array_values) < array_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    actual_values = decoder(stream)
    assert actual_values == array_values[:array_size]


@pytest.mark.parametrize(
    'types,data,expected',
    (
        (
            ('address', 'uint256'),
            (
                '0x'
                '000000000000000000000000abf7d8b5c1322b3e553d2fac90ff006c30f1b875'
                '0000000000000000000000000000000000000000000000000000005d21dba000'
            ),
            ('0xabf7d8b5c1322b3e553d2fac90ff006c30f1b875', 400000000000)
        ),
        (
            ('uint256', 'bytes'),
            (
                '0x'
                '0000000000000000000000000000000000000000000000000000000000000000'
                '0000000000000000000000000000000000000000000000000000000000000040'
                '0000000000000000000000000000000000000000000000000000000000000000'
                '0000000000000000000000000000000000000000000000000000000000000000'
            ),
            (0, b''),
        ),
    ),
)
def test_tuple_decoder(types, data, expected):
    decoders = [registry.get_decoder(t) for t in types]
    decoder = TupleDecoder(decoders=decoders)
    stream = ContextFramesBytesIO(decode_hex(data))
    actual = decoder(stream)
    assert actual == expected


@settings(max_examples=250)
@given(
    high_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    low_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_unsigned_real(high_bit_size,
                              low_bit_size,
                              integer_bit_size,
                              stream_bytes,
                              data_byte_size):
    if integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedRealDecoder(
                value_bit_size=integer_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif high_bit_size + low_bit_size != integer_bit_size:
        with pytest.raises(ValueError):
            UnsignedRealDecoder(
                value_bit_size=integer_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = UnsignedRealDecoder(
            value_bit_size=integer_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    padding_bytes = stream_bytes[:data_byte_size][:data_byte_size - integer_bit_size // 8]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    unsigned_integer_value = big_endian_to_int(stream_bytes[:data_byte_size])

    with decimal.localcontext(abi_decimal_context):
        raw_real_value = decimal.Decimal(unsigned_integer_value) / 2 ** low_bit_size

    actual_value = quantize_value(raw_real_value, low_bit_size)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    high_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    low_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
@example(
    high_bit_size=8,
    low_bit_size=8,
    integer_bit_size=16,
    stream_bytes=b'\xff\xff\xff\xa9\xf5\xb3',
    data_byte_size=3,
)
def test_decode_signed_real(high_bit_size,
                            low_bit_size,
                            integer_bit_size,
                            stream_bytes,
                            data_byte_size):
    if integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedRealDecoder(
                value_bit_size=integer_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif high_bit_size + low_bit_size != integer_bit_size:
        with pytest.raises(ValueError):
            SignedRealDecoder(
                value_bit_size=integer_bit_size,
                high_bit_size=high_bit_size,
                low_bit_size=low_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = SignedRealDecoder(
            value_bit_size=integer_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)

    padding_offset = data_byte_size - integer_bit_size // 8
    data_offset = padding_offset + integer_bit_size // 8

    padding_bytes = stream_bytes[:data_byte_size][:padding_offset]
    data_bytes = stream_bytes[:data_byte_size][padding_offset:data_offset]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif not is_valid_padding_bytes(padding_bytes, data_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    if padding_bytes:
        if decoded_value >= 0:
            assert bytes(set(padding_bytes)) == b'\x00'
        else:
            assert bytes(set(padding_bytes)) == b'\xff'

    _, upper_bound = compute_signed_integer_bounds(high_bit_size + low_bit_size)

    unsigned_integer_value = big_endian_to_int(data_bytes)
    if unsigned_integer_value >= upper_bound:
        signed_integer_value = unsigned_integer_value - 2 ** (high_bit_size + low_bit_size)
    else:
        signed_integer_value = unsigned_integer_value

    with decimal.localcontext(abi_decimal_context):
        raw_actual_value = decimal.Decimal(signed_integer_value) / 2 ** low_bit_size

    actual_value = quantize_value(raw_actual_value, low_bit_size)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_unsigned_fixed(value_bit_size,
                              frac_places,
                              stream_bytes,
                              data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedFixedDecoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    decoder = UnsignedFixedDecoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    stream = ContextFramesBytesIO(stream_bytes)
    padding_bytes = stream_bytes[:data_byte_size][:data_byte_size - value_bit_size // 8]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    if is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    # Ensure no exceptions
    actual_value = decoder(stream)


@settings(max_examples=250)
@given(
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_signed_fixed(value_bit_size,
                             frac_places,
                             stream_bytes,
                             data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedFixedDecoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    decoder = SignedFixedDecoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    stream = ContextFramesBytesIO(stream_bytes)

    padding_offset = data_byte_size - value_bit_size // 8
    data_offset = padding_offset + value_bit_size // 8

    padding_bytes = stream_bytes[:data_byte_size][:padding_offset]
    data_bytes = stream_bytes[:data_byte_size][padding_offset:data_offset]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    if not is_valid_padding_bytes(padding_bytes, data_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    actual_value = decoder(stream)

    if padding_bytes:
        if actual_value >= 0:
            assert bytes(set(padding_bytes)) == b'\x00'
        else:
            assert bytes(set(padding_bytes)) == b'\xff'
